/** dslconf - Simple DSL modem configuration protocol. */
#ifndef _DSLCONF_H
#define _DSLCONF_H

#include <semaphore.h>	/* sem_t */

/** dslconf header format */
struct dslconf_hdr {
	__u8 type;		/**< see enum dc_type */
	__u8 flags;		/**< FLAG_{REQ,ACK,NACK} */
	__u32 dlen;		/**< length of data (i.e. full packet payload) */
} __attribute__((__packed__));

/** dslconf packet format, for easier mapping into buffers of various size */
struct dslconf_pkt {
	struct dslconf_hdr hdr;	/**< packet header */
	char data[1];		/**< start of packet payload */
} __attribute__((__packed__));

/** dslconf request types */
enum dc_type {
	/** NOP packet, reply is always ACK
	 * request data: none
	 * reply data: none
	 * NACK on: never */
	REQ_NOP = 0,

	/** EXEC packet: call a command in shell
	 * request data: <command> [<arg1> ...]
	 * reply data: command output (if any)
	 * NACK on: popen() fails */
	REQ_EXEC = 1,

	/** GET_ENV: call getenv() on server side
	 * request data: <variable_name>
	 * reply data: <value>
	 * NACK on: variable undefined */
	REQ_GET_ENV = 2,

	/** SET_ENV: call setenv() on server side
	 * request data: <variable_name>=<value>
	 * reply data: none
	 * NACK on: invalid syntax, setenv() fails */
	REQ_SET_ENV = 3,

	/** FLASH_CFG: write a new config to flash (ar7 only)
	 * request data: <config data in binary form>
	 * reply data: useless success message
	 * NACK on: write to device node fails */
	REQ_FLASH_CFG = 4,

	/** REEXEC: kill and reexecute a command
	 * request data: <command> [<arg1> ...]
	 * reply data: none
	 * NACK on: internal pidof() failed, kill() failed */
	REQ_REEXEC = 5,

	/** GET_GW: request various gateway information
	 * request data: (iface | ip | count)
	 * reply data: (<default route's interface> | <gateway IP> | <number of default routes>)
	 * NACK on: internal data retrieval fails, illegal syntax */
	REQ_GET_GW = 6,

	/** GET_IP: return the IP of an interface (or the gateway interface)
	 * request data: (<ifname> | gwiface)
	 * reply data: <IP in dotted decimal format>
	 * NACK on: never */
	REQ_GET_IP = 7,

	/** SET_IP: set the IP of an interface
	 * request data: <ifname> <IP>(/<prefix> | <netmask>)
	 * reply data: none
	 * NACK on: syntax error, ioctl() fails */
	REQ_SET_IP = 8,

	/** START_LOGIC: trigger the modem's dial-in logic (NOP if already running) (ar7 only)
	 * request data: none
	 * reply data: none
	 * NACK on: internal error */
	REQ_START_LOGIC = 9,

	/** RELOAD_LOGIC: make the logic reapply the config from flash
	 * (implicitly calls START_LOGIC if not yet running)
	 * request data: none
	 * reply data: none
	 * NACK on: no logic running yet */
	REQ_RELOAD_LOGIC = 10,

	/** FLASH_FW: flash a new firmware image (ar7 only)
	 * request data: url to firmware *upgrade* image
	 * reply data: none
	 * NACK on: never */
	REQ_FLASH_FW = 11,

	/** GET_PUB_IP: retrieve the public IP 'stealip' has
	 * stolen from the ppp interface before.
	 * request data: none
	 * reply data: <PubIP>/<Prefix> <GW-IP>
	 * NACK on: not available */
	REQ_GET_PUB_IP = 12,

	/** SYNC_STAT: request the modem's synchronisation state
	 * reply data: (IDLE | INIT | SHOWTIME | QUIET) <attempt num>
	 *   IDLE: modem has no sync (cable disconnected)
	 *   INIT: modem is synchronising right now (cable just plugged in)
	 *   SHOWTIME: modem is in sync with peer
	 *   QUIET: modem was in sync but line is dead now (cable just pulled)
	 *   <attempt num>: modem internal counter for sync attempts (wraps after 29)
	 * NACK on: non-existing file, unexpected content */
	REQ_SYNC_STAT = 13,

	/** RATE_INFO: get the available US/DS bitrates
	 * reply data: <US rate> <DS rate>
	 * NACK on: non-existing file, unexpected content */
	REQ_RATE_INFO = 14,

	/** PPP_INFO: retrieve the PPP connection status
	 * reply data: some nice status message:
	 *	"Unknown"
	 *	"Authenication Failure with Peer"
	 *	"Connection Successfully Established"
	 *	"Connect Delayed"
	 * NACK on: never */
	REQ_PPP_INFO = 15,

	/** STOP_LOGIC: kill any active dial-in process (NOP if not running)
	 * request data: none
	 * reply data: none
	 * NACK on: never */
	REQ_STOP_LOGIC = 16,

	/** GET_VERSION: request the server implementation's version
	 * request data: none
	 * reply data: version string
	 * NACK on: earlier versions not having this implemented */
	REQ_GET_VERSION = 17,

	/** FS_STAMP: return the filesystem build timestamp
	 * request data: none
	 * reply data: timestamp in format YYYYMMDDHHmmSS
	 * NACK on: earlier versions not having this implemented */
	REQ_FS_STAMP = 18,

	/** GET_PPPD_LOG: print the pppd log
	 * request data: none
	 * reply data: pppd logfile contents
	 * NACK on: earlier versions not having this implemented */
	REQ_GET_PPPD_LOG = 19,

	/** SET_VDSL_CFG: write /etc/vdsl.conf (fusiv only)
	 * request data: <config content in ASCII form>
	 * reply data: none
	 * NACK on: config writing fails */
	REQ_SET_VDSL_CFG = 20,

	/** USE_BUILTIN_VDSL_CFG: use one of /etc/vdsl.conf.* (fusiv only)
	 * request data: (annexa|annexb|vdsl)
	 * reply data: none
	 * NACK on: setting config fails */
	REQ_USE_BUILTIN_VDSL_CFG = 21,

	/** WAN_MGR: start/stop WAN manager (fusiv only)
	 * request data: <parameters matching wan_mgr.sh>
	 * reply data: <wan_mgr.sh output (if any)>
	 * NACK on: calling wan_mgr.sh fails */
	REQ_WAN_MGR = 22,

	/** VDSL_CFG_SHA1: get vdsl config SHA1 sum (fusiv only)
	 * request data: [annexa|annexb]
	 * reply data: checksum as 40 chars hex string of either builtin
	 *             config or the currently active one if no params were passed */
	REQ_VDSL_CFG_SHA1 = 23,

	/** VDSL_MON_STATS: get whole vdsl mon stats in binary format (fusiv only)
	 * request data: none
	 * reply data: packed vdsl_mon_stats data in network byte order */
	REQ_VDSL_MON_STATS = 24,

	/** SCMVER: return the BSP's git commit ID (usually tag)
	 * request data: none
	 * reply data: any meaningful string
	 * NACK on: earlier versions not having this implemented, file access fail */
	REQ_SCMVER = 25,

	/** NOTIFY, WARNING, ERRROR: no real request types, used only as reply
	 *                           to communicate server messages to the client
	 * request data: not applicable
	 * reply data: server message of given severity */
	REQ_NOTIFY = 253,
	REQ_WARNING = 254,
	REQ_ERROR = 255,
};

/** valid flag bits in dslconf_hdr.flags */
#define FLAG_REQ	(1 << 0)	/**< packet is a request from client to server */
#define FLAG_ACK	(1 << 1)	/**< packet is an ACK (positive reply) from server */
#define FLAG_NACK	(1 << 2)	/**< packet is a NACK (negative reply) from server */

/** Association between dc_type value and handler routine on server side. */
struct listener_callback {
	enum dc_type type;
	int (*callback)(struct interface, struct dslconf_pkt *);
};

/** allocate a dslconf packet */
struct dslconf_pkt *alloc_dslconf_pkt(__u8, __u8, const void *, size_t);

/** receive a complete and valid dslconf packet */
ssize_t recv_dslconf_packet(struct interface, void **, __u32 *);

/** client side transaction: send a request and wait for the appropriate reply */
int do_transaction(struct interface, struct dslconf_pkt *, struct dslconf_pkt **);

/** server side listening: waits for request and calls the appropriate callback */
void listener(struct interface, const struct listener_callback *, sem_t *);

/** simple interface for server replies */
int send_reply_pkt(struct interface iface, __u8 type, __u8 flags,
		const void *data, __u32 datalen);

/* send_reply_msg:
 * By convention, although this routine is used to reply a string, the final
 * binary zero is not sent out. So a receiver intending to print the reply
 * message has to manually add it based on the length field.
 * XXX: use va_list here instead of just const char *msg? */
#define send_reply_msg(i, t, f, m) \
				send_reply_pkt(i, t, f, m, strlen(m))
#define send_ack_msg(i, t, m)	send_reply_msg(i, t, FLAG_ACK, m)
#define send_nack_msg(i, t, m)	send_reply_msg(i, t, FLAG_NACK, m)
#define send_nack_errno(l, t)	send_nack_msg(l, t, strerror(errno))
#define send_ack(i, t)		send_ack_msg(i, t, "")
#define send_nack(i, t)		send_nack_msg(i, t, "")

#endif /* _DSLCONF_H */
