#define _POSIX_C_SOURCE 201309L
#include <errno.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "simplecom.h"
#include "dslconf.h"
#include "pidof.h"
#include "server.h"

#define UINT32 uint32_t
#define INT32 int32_t

typedef struct
{
    /* output from cmd 11 */
    char   LineTypeDesc[40];  /* Line Type Description */
    UINT32 LineType;  /* Line Type */
    UINT32 UsLineRate; /* Upstream Line Rate */
    UINT32 DsLineRate; /* Downstream Line Rate */
    UINT32 DsAttainableRate; /* DS attainable data rate */
    UINT32 UsAttainableRate; /* US attainable data rate */
    short  txPwr; /* Tx Total Power (Aggregate Power) */
    short  txPwr_FE; /* Far-end Tx Total Power */

    /* output from cmd 13 */
    INT32  SnrMargin;   /* SNR Margin */
    UINT32 Attn;    /* Attenuation */
    INT32  SnrAvgMargin;  /* Average SNR Margin */
    INT32  SnrAvg;    /* Average SNR */

    /* Farend Line Attenuation, SNR margin and US attainable datarate */
    float  FeAttn[5];  /* Line Attenuation */
    float  FeSNRM[6];  /* SNR Margin */
    unsigned char FeAttnStatus;  /* Status Available/not Avialable */
    unsigned char FeSNRMStatus;  /* Status Available/not Avialable */

    /* output from cmd 12 */
    UINT32 NeHEC_F;  /* Near-end Header Error Control (HEC) Fast count */
    UINT32 NeHEC_S;  /* Near-end Header Error Control (HEC) Slow count */
    UINT32 FeHEC_F;  /* Far-end Header Error Control (HEC) Fast count */
    UINT32 FeHEC_S;  /* Far-end Header Error Control (HEC) Slow count */
    UINT32 NeCRC_F;  /* Near-end Cyclic Redundancy Check (CRC) Fast count */
    UINT32 NeCRC_S;  /* Near-end Cyclic Redundancy Check (CRC) Slow count */
    UINT32 FeCRC_F;  /* Far-end Block Error (BE) Fast count */
    UINT32 FeCRC_S;  /* Far-end Block Error (BE) Slow count */
    /* 
     * Commenting the below members as these are not used anywhere; when this 
     * is uncommented, please sync fusiv_src/user/mini_httpd-1.19/vdslmon.h;
     * Otherwise "xdslmon" will always report "xDSL port is not in showtime"
     * irrespective of the actual status.
     */
    //INT32 ShowAttnPayRate; /*Showtime attainable payload rate */
    //INT32 ShowAttnLineRate; /*Showtime attainable line rate */
} vdsl_mon_stats_t;

static vdsl_mon_stats_t vdsl_mon_stats;
static bool vdsl_mon_stats_valid = false;

#define VDSLCMD_OUT		"/tmp/saved_vdslcmd.out"
#define VDSLCMD_BOND_OUT	"/tmp/saved_vdslcmd_bonding.out"
static void fetch_vdsl_mon_stats(void)
{
	int fd;

	fd = open(VDSLCMD_BOND_OUT, O_RDONLY);
	if (fd == -1)
		fd = open(VDSLCMD_OUT, O_RDONLY);

	if (fd == -1 ||
	    (read(fd, (void *)&vdsl_mon_stats, sizeof(vdsl_mon_stats)) == -1))
		vdsl_mon_stats_valid = false;
	else
		vdsl_mon_stats_valid = true;
	close(fd);
}

static int sync_stat_update(void)
{
	static char lt_desc[40] = {0};

	fetch_vdsl_mon_stats();
	if (vdsl_mon_stats_valid) {
		if (strcmp(lt_desc, vdsl_mon_stats.LineTypeDesc)) {
			memcpy(lt_desc, vdsl_mon_stats.LineTypeDesc, 40);
			return 1;
		}
	}
	return 0;
}

static void sync_stat_notify(struct interface iface)
{
	send_ack_msg(iface, REQ_SYNC_STAT, vdsl_mon_stats.LineTypeDesc);
}

static int sync_stat_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char lt_desc[40] = {0};
	if (vdsl_mon_stats_valid) {
		sscanf(vdsl_mon_stats.LineTypeDesc, "%[^\n]", lt_desc);
		return send_ack_msg(iface, pkt->hdr.type, lt_desc);
	}
	return send_nack_msg(iface, pkt->hdr.type, "invalid stats");
}

DSLCONF_CMD(SYNC_STAT, sync_stat);

static int rate_info_update(void)
{
	static uint32_t us_rate = 0, ds_rate = 0;

	if (!vdsl_mon_stats_valid)
		return 0;

	if (us_rate != vdsl_mon_stats.UsLineRate ||
	    ds_rate != vdsl_mon_stats.DsLineRate) {
		us_rate = vdsl_mon_stats.UsLineRate;
		ds_rate = vdsl_mon_stats.DsLineRate;
		return 1;
	}
	return 0;
}

static void rate_info_notify(struct interface iface)
{
	char buf[256];

	snprintf(buf, 256, "%d %d",
	         vdsl_mon_stats.UsLineRate,
	         vdsl_mon_stats.DsLineRate);
	send_ack_msg(iface, REQ_RATE_INFO, buf);
}

static int rate_info_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char buf[256];

	if (!vdsl_mon_stats_valid)
		return send_nack_msg(iface, pkt->hdr.type, "stats not valid");

	if (!strcmp(vdsl_mon_stats.LineTypeDesc, "Port is not in showtime."))
		return send_ack_msg(iface, pkt->hdr.type, "0 0");

	snprintf(buf, 256, "%d %d",
	         vdsl_mon_stats.UsLineRate,
	         vdsl_mon_stats.DsLineRate);
	return send_ack_msg(iface, pkt->hdr.type, buf);
}

DSLCONF_CMD(RATE_INFO, rate_info);

static float htonf(float val)
{
	union {
		float f;
		uint32_t d;
	} tmp = { .f = val };
	tmp.d = htonl(tmp.d);
	return tmp.f;
}

static int vdsl_mon_stats_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	struct {
	    char LineTypeDesc[40];
	    UINT32 LineType, UsLineRate, DsLineRate,
		   DsAttainableRate, UsAttainableRate;
	    short txPwr, txPwr_FE;

	    INT32  SnrMargin;
	    UINT32 Attn;
	    INT32  SnrAvgMargin, SnrAvg;

	    float  FeAttn[5], FeSNRM[6];
	    unsigned char FeAttnStatus, FeSNRMStatus;

	    UINT32 NeHEC_F, NeHEC_S, FeHEC_F, FeHEC_S,
		   NeCRC_F, NeCRC_S, FeCRC_F, FeCRC_S;
	} __attribute__((packed)) stats_packed;
	int i;

	fetch_vdsl_mon_stats();
	if (!vdsl_mon_stats_valid)
		return send_nack_msg(iface, pkt->hdr.type, "stats not valid");

#if defined(DEBUG)
#define PD(f) printf(#f ": %d\n", vdsl_mon_stats.f)
#define PF(f) printf(#f ": %f\n", vdsl_mon_stats.f)
#else
#define PD(f) /* nothing */
#define PF(f) /* nothing */
#endif
#define HTONL(f) PD(f); stats_packed.f = htonl(vdsl_mon_stats.f)
#define HTONS(f) PD(f); stats_packed.f = htons(vdsl_mon_stats.f)
#define HTONF(f) PF(f); stats_packed.f = htonf(vdsl_mon_stats.f)
#define HTONB(f) PD(f); stats_packed.f = vdsl_mon_stats.f
	memcpy(stats_packed.LineTypeDesc, vdsl_mon_stats.LineTypeDesc, 40);
	HTONL(LineType);
	HTONL(UsLineRate);
	HTONL(DsLineRate);
	HTONL(DsAttainableRate);
	HTONL(UsAttainableRate);
	HTONS(txPwr);
	HTONS(txPwr_FE);
	HTONL(SnrMargin);
	HTONL(Attn);
	HTONL(SnrAvgMargin);
	HTONL(SnrAvg);
	for (i = 0; i < 5; i++) {
		HTONF(FeAttn[i]);
		HTONF(FeSNRM[i]);
	}
	HTONF(FeSNRM[5]);
	HTONB(FeAttnStatus);
	HTONB(FeSNRMStatus);
	HTONL(NeHEC_F);
	HTONL(NeHEC_S);
	HTONL(FeHEC_F);
	HTONL(FeHEC_S);
	HTONL(NeCRC_F);
	HTONL(NeCRC_S);
	HTONL(FeCRC_F);
	HTONL(FeCRC_S);

	return send_reply_pkt(iface, pkt->hdr.type, FLAG_ACK,
			&stats_packed, sizeof(stats_packed));
}
SIMPLE_DSLCONF_CMD(VDSL_MON_STATS, vdsl_mon_stats);

#define VDSL_CFG "/etc/vdsl.conf"
#define VDSL_BIN_CFG "/etc/config/dslSavedConfig.conf"
static int set_vdsl_cfg_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	int fd;

	unlink(VDSL_CFG);
	unlink(VDSL_BIN_CFG);

	if (pkt->hdr.type == REQ_USE_BUILTIN_VDSL_CFG) {
		char oldpath[256];

		snprintf(oldpath, 256, VDSL_CFG ".%s", pkt->data);

		if (access(oldpath, F_OK) || symlink(oldpath, VDSL_CFG))
			return send_nack_errno(iface, pkt->hdr.type);

		return !(!send_ack(iface, pkt->hdr.type) && !system("reboot"));
	}

	if ((fd = open(VDSL_CFG, O_CREAT | O_TRUNC | O_WRONLY, 00644)) == -1)
		return send_nack_msg(iface, pkt->hdr.type, "could not create " VDSL_CFG);

	if (write(fd, pkt->data, pkt->hdr.dlen) != (signed)pkt->hdr.dlen) {
		close(fd);
		return send_nack_msg(iface, pkt->hdr.type, "could not write " VDSL_CFG);
	}

	close(fd);
	return !(!send_ack(iface, pkt->hdr.type) && !system("reboot"));
}
SIMPLE_DSLCONF_CMD(SET_VDSL_CFG, set_vdsl_cfg);
SIMPLE_DSLCONF_CMD(USE_BUILTIN_VDSL_CFG, set_vdsl_cfg);

static char ppp_status[128];

static int ppp_info_update(void)
{
	char *buf, *p;

	if (!(buf = get_cmd_output("/usr/sbin/ppp_status.sh"))) {
		derror("failed to get cmd output from /usr/sbin/ppp_status.sh");
		return -1;
	}

	/* drop at least the trailing newline character */
	p = buf + strlen(buf) - 1;
	while (*p < 33 || *p > 126)
		*(p--) = 0;

	if (strcmp(ppp_status, buf)) {
		sprintf(ppp_status, "%s", buf);
		free(buf);
		return 1;
	}
	free(buf);
	return 0;
}

static void ppp_info_notify(struct interface iface)
{
	send_ack_msg(iface, REQ_PPP_INFO, ppp_status);
}

static int ppp_info_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	return send_ack_msg(iface, pkt->hdr.type, ppp_status);
}

DSLCONF_CMD(PPP_INFO, ppp_info);

#define WAN_MGR_PATH "/usr/sbin/wan_mgr.sh"
static int wan_mgr_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char cmd[256], *output;

	/* stop wan_mgr first, just in case ... */
	snprintf(cmd, 256, "%s stop", WAN_MGR_PATH);
	if (system(cmd) < 0)
		return send_nack_msg(iface, pkt->hdr.type,
				"stopping wan_mgr failed");

	snprintf(cmd, 256, "%s %s", WAN_MGR_PATH, pkt->data);

	dnotify("wan_mgr callback being called, executing '%s'", cmd);

	if (!(output = get_cmd_output(cmd)))
		return send_nack_errno(iface, pkt->hdr.type);

	send_ack_msg(iface, pkt->hdr.type, output);

	free(output);
	return 0;
}
SIMPLE_DSLCONF_CMD(WAN_MGR, wan_mgr);

#define SHA1SUM_PATH "/usr/bin/sha1sum"
static int vdsl_cfg_sha1_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char cmd[1024], *output;

	snprintf(cmd, 1024, "%s %s%s%s", SHA1SUM_PATH, VDSL_CFG,
			pkt->data[0] ? "." : "", pkt->data);
	if (!(output = get_cmd_output(cmd))) {
		send_nack_errno(iface, pkt->hdr.type);
		return 0;
	}

	/* the SHA1 checksum consists of 40 characters */
	output[40] = 0;

	send_ack_msg(iface, pkt->hdr.type, output);

	free(output);
	return 0;
}
SIMPLE_DSLCONF_CMD(VDSL_CFG_SHA1, vdsl_cfg_sha1);
